@@ -4,7 +4,7 @@
 #
 # Script to clean the apt-cacher cache.
 #
-# Copyright (C) 2007-14, Mark Hindley <mark@hindley.org.uk>
+# Copyright (C) 2007-16, Mark Hindley <mark@hindley.org.uk>
 # Copyright (C) 2005, Eduard Bloch <blade@debian.org>
 # Copyright (C) 2002-03, Jonathan Oxer <jon@debian.org>
 # Portions  (C) 2002, Jacob Lundberg <jacob@chaos2.org>
@@ -15,7 +15,7 @@ use strict;
 use warnings;
 use lib '/usr/share/apt-cacher/lib';
 
-use sigtrap qw(die normal-signals);
+use sigtrap qw(die normal-signals error-signals);
 use Cwd ();
 use Fcntl qw/:DEFAULT :flock F_SETFD/;
 use Getopt::Long qw(:config no_ignore_case);
@@ -167,17 +167,9 @@ sub open_lock {
 }
 
 sub get {
-    my ($file, $use_url) = @_;
-    my $path_info;
-    # use path if stored in cached header/complete file
-    if($use_url && (my $url = get_original_url($file))) {
-	$path_info= $url;
-    }
-    else {
-	$path_info=$file;
-	$path_info=~s/^/\//;
-	$path_info=~s/_/\//g;
-    }
+    my ($file) = @_;
+
+    my $path_info = get_original_url($file);
     defined(my $refresh_pid = open(my $fh, '|-')) || die "Failed to open refresh pipe: $!";
     if ($refresh_pid){
 	printmsg "Get $path_info\n";
@@ -210,7 +202,13 @@ sub pdiff {
 	printmsg "Upstream repository for $name not standard hierarchy, skipping attempting to patch\n";
 	return;
     }
-    my ($basename,$type) = ($name =~ /(^.+?)(\.(?:bz2|gz))?$/);
+
+    if ($name =~ /\.xz$/ && !eval{require IO::Compress::Xz}) {
+	printmsg "Install libio-compress-lzma-perl to enable Xz pdiff support\n";
+	return;
+    }
+
+    my ($basename,$type) = ($name =~ /(^.+?)(\.(?:bz2|(?:x|g)z))?$/);
     (my $release = $basename) =~ s/(?:main|contrib|non-free).*$/{In,}Release/;
     (my $diffindex = $basename) .= '.diff_Index';
 
@@ -323,28 +321,40 @@ sub pdiff {
 	}
     }
     seek($diffin_fh,0,0) || die "Seek failed: $!"; # rewind
-    my $curr= <$diffin_fh>; # read first line
-    chomp $curr; # remove trailing \n
-#    printmsg "$diffindex: $curr\n";
-    my ($target_sha1, $target_size) = (split (/\s+/,$curr))[1,2];
-    if ($digest eq $target_sha1) { # check this matches /SHA1/
-	printmsg "SHA1 match: $name already up to date\n";
-	_flock($diffin_fh, LOCK_UN);
-	close ($diffin_fh);
-	return 1; # success
-    }
-    else {
-	while (<$diffin_fh>) {
-	    next if (/^SHA1-History:/); # skip header
-	    last if (/^SHA1-Patches:/);# end of history
-	    push @hist, $_;
-	    next;
-	}
-	while (<$diffin_fh>) {
-	    push @patch, $_; # To EOF
-	    next;
-	}
-    }
+    my ($header, $target_sha1, $target_size);
+    while (<$diffin_fh>) {; # read first line
+			  chomp; # remove trailing \n
+			  #    printmsg "$diffindex: $_\n";
+			  # Store SHA1 only
+			  /^SHA1-Current:\s+([[:xdigit:]]+)\s+(\d+)/ && do {
+			      $target_sha1 = $1;
+			      $target_size = $2;
+			      if ($digest eq $target_sha1) { # check this matches /SHA1/
+				  printmsg "SHA1 match: $name already up to date\n";
+				  _flock($diffin_fh, LOCK_UN);
+				  close ($diffin_fh);
+				  return 1; # success
+			      }
+			      next;
+			  };
+			  /^SHA\d+-\S+:$/ && do {
+			      $header = $_;
+			      next;
+			  };
+			  /^\s+[[:xdigit:]]+\s+\d+\s+\S+/ && do {
+			      if (!defined $header) {
+				  warn "$diffindex parse error: undefined section.\n"
+			      }
+			      if ($header eq 'SHA1-History:') {
+				  push @hist, $_;
+			      }
+			      elsif ($header eq 'SHA1-Patches:') {
+				  push @patch, $_;
+			      }
+			      next;
+			  };
+		      }
+    undef $header;
     _flock($diffin_fh, LOCK_UN);
     close ($diffin_fh);
 
@@ -370,8 +380,8 @@ sub pdiff {
 
     open(my $diffs_fh, ">", \$diffs) || die "Failed to open in memory diff file: $!";
     for (@patch[$diff .. $#patch]) {
-	my ($pdiffsha1, $size, $suff) = split;
-	my $pdiff = "$basename.diff_$suff.gz";
+	my ($pdiffsha1, $size, $id) = split;
+	my $pdiff = "$basename.diff_$id.gz";
 	if (!-f $pdiff) {
 	    if (!$offline) {
 		get($pdiff);
@@ -419,8 +429,8 @@ sub pdiff {
     print $patchpipe "w\n"; # ed write command
     close($patchpipe);
     chdir $cwd or die "Unable to restore working directory: $!"; # Restore
-    my $rstat =($? >> 8);
-    if ($rstat) {
+    if ($? > 0) {
+	my $rstat =($? >> 8);
 	warn "Patching failed (exit code $rstat), aborting\n";
 	return;
     }
@@ -444,9 +454,11 @@ sub pdiff {
 	    seek($cfh, 0, 0) || die "Seek failed: $!";
 	    my ($z,$encoding) = ($name=~/bz2$/ ?
 				 ((IO::Compress::Bzip2->new($cfh)), "x-bzip2") :
-				 ($name=~/gz$/ ?
-				  ((IO::Compress::Gzip->new($cfh, -Level => 9)), "x-gzip") :
-				  $cfh));
+				 ($name=~/xz$/ ?
+				  ((IO::Compress::Xz->new($cfh)), "x-xz") :
+				  ($name=~/gz$/ ?
+				   ((IO::Compress::Gzip->new($cfh, -Level => 9)), "x-gzip") :
+				   $cfh)));
 
 	    while (<$tfh>) {
 		$z->print($_);
@@ -478,26 +490,54 @@ sub pdiff {
 }
 
 # Calls _db_compact to do the work and reports results
-# Arg: DB handle ref
 sub db_compact {
-    my ($dbh) = @_;
+    my $dbh = db();
+    printmsg "Waiting for CDS lock...\n";
+    my $db_lock = $dbh->cds_lock();
     printmsg "Compacting checksum database....\n";
-    while (my ($status, %results) = @{_db_compact($dbh)}) {
-	if ($status) {
-	    printmsg "db_compact failed: $status\n";
-	    last;
-	}
-	else {
-	    printmsg " Compacted ". $results{compact_pages_free} ." pages\n Freed ". $results{compact_pages_truncated} ." pages\n";
-	    if ($results{compact_pages_free} + $results{compact_pages_truncated} == 0) {
-		printmsg "Done!\n";
-		last;
-	    }
-	}
+    my ($status, %results) = @{_db_compact($dbh)};
+    if ($status) {
+	printmsg "db_compact failed: $status\n";
+    }
+    else {
+	printmsg " Compacted ". $results{compact_pages_free} ." pages\n Freed ". $results{compact_pages_truncated} ." pages\n";
     }
     return;
 }
 
+# Parse InRelease for by-hash data
+# Return arrayref of relevant (Sources, Packages, diff_Index) by-hash filenames
+sub by_hash {
+    my ($file) = @_;
+    my @ret;
+
+    open(my $fh, '<', $file) or
+      die "Failed to open $file: $!";
+
+    while (<$fh>) {
+	# Loop through file and only parse it fully if Acquire-By-Hash is set
+	/^Acquire-By-Hash: yes$/ && do {
+	    my %parsed;
+	    extract_sums($file, $fh, \%parsed) || die("Error extracting by-hash data from $file\n");
+	    foreach my $index_file (keys %parsed) {
+		next unless $index_file =~ /_(?:Sourc|Packag)es(?:\.(?:gz|xz|bz2|diff_Index))?/;
+		my $href = hashify(\$parsed{$index_file});
+		while (my ($k,$v) = each %$href) {
+		    next if $k eq 'size';
+		    $k = uc $k;
+		    $k .= 'Sum' if $k eq 'MD5';
+		    if ((my $possible = $index_file) =~ s/[^_]+$/by-hash_${k}_$v/) {
+			push (@ret,  $possible);
+		    }
+		}
+	    }
+	    last;
+	}
+    }
+    return \@ret;
+}
+
+
 #############################################################################
 # Manipulate checksum database
 if (@db_mode || $db_recover){
@@ -513,6 +553,10 @@ if (@db_mode || $db_recover){
     $verbose = 1; # Just for now
 
     if ($db_recover) {
+	if (my $env=env(1)) { # No locking
+	    printmsg "Set Panic on old database environment...";
+	    $env->set_flags(BerkeleyDB->DB_PANIC_ENVIRONMENT, 1);
+	}
 	printmsg "Running database recovery...";
 	db_recover();
 	printmsg "Done!\n";
@@ -523,7 +567,7 @@ if (@db_mode || $db_recover){
   SWITCH:
     while (local $_ = shift @db_mode) {
 	/^import$/ && do {
-	    foreach (glob('*es.bz2 *es.gz *es *Release *diff_Index')) {
+	    foreach (glob('*es.bz2 *es.xz *es.gz *es *Release *diff_Index')) {
 		open(my $fh, '<', $_)|| do {
 		    warn "Failed to open $_ for import: $!";
 		    next;
@@ -534,7 +578,7 @@ if (@db_mode || $db_recover){
 	    next SWITCH;
 	};
 	/^compact$/ && do {
-	    db_compact(db());
+	    db_compact();
 	    next SWITCH;
 	};
 	/^(?:dump|search)$/ && do {
@@ -544,7 +588,7 @@ if (@db_mode || $db_recover){
 		die "No search expression given\n" if !$re;
 		die "Invalid character '$1' in search\n" if $re =~ /([^$ok_chars])/o; # sanitize
 	    }
-	    my $cursor = get_cursor(db());
+	    my $cursor = db_cursor();
 	    my ($filename,$data) = ('','');
 	    while (cursor_next($cursor, \$filename, \$data) == 0)
 	      {
@@ -562,7 +606,7 @@ if (@db_mode || $db_recover){
 	    my $re = shift @db_mode;
 	    die "No give regex to match files to delete\n" if !$re;
 	    die "Invalid character '$1' in pattern\n" if $re =~ /([^$ok_chars])/o; # sanitize
-	    my $cursor = get_cursor(db(),1);
+	    my $cursor = db_cursor(1);
 	    my ($filename,$data) = ('','');
 	    while (cursor_next($cursor, \$filename, \$data) == 0)
 	      {
@@ -573,21 +617,23 @@ if (@db_mode || $db_recover){
 	    next SWITCH;
 	};
 	/^failcheck$/ && do {
-	    printmsg 'Connecting to database....';
-	    # Just connect to the database which runs failchk()
-	    if (db(1)) { # Without locking
+	    printmsg 'Connecting to database environment....';
+	    # Just connect to the database environment which runs failchk()
+	    if (env(1)) { # Without locking
 		printmsg "Success!\n";
 	    }
 	    next SWITCH;
 	};
 	/^verify$/ && do {
-	    printmsg "Waiting for exclusive lock...";
-	    if (db_flock(LOCK_EX)){
+	    my $db=db();
+	    printmsg "Waiting for CDS lock...";
+	    if (my $db_lock = $db->cds_lock){
+		$db->db_sync() == 0 || warn "db_sync failed:  $BerkeleyDB::Error";
 		printmsg "Got it!\nVerifying database...";
-		printmsg db_verify("$cfg->{cache_dir}/sums.db", temp_env()) ? "Failed! " . db_error() . "\n" : "Passed!\n";
+		printmsg db_verify("$cfg->{cache_dir}/sums.db", $db->Env) ? "Failed! " . db_error() . "\n" : "Passed!\n";
 	    }
 	    else {
-		warn "Unable to get exclusive database lock: $!\n";
+		warn "Unable to get CDS database lock: $!\n";
 	    }
 	    next SWITCH;
 	};
@@ -615,10 +661,11 @@ if (defined $cfg->{offline_mode} && $cfg->{offline_mode}) {
 	$offline = 1;
 }
 
-use GDBM_File;
+use DB_File ();
 open(my $tmpfile, "+>", undef) or die $!;
-tie my %valid, 'GDBM_File', fd_path($tmpfile), &GDBM_NEWDB|&GDBM_FAST, oct(600) # Does a separate open
-  or die "GDBM_File tie failed: $!";
+chmod oct(600), $tmpfile unless (stat $tmpfile)[2] & oct(7777) ^ oct(600); # workaround PerlIO_tmpfile() incorrect mode in 5.22.1
+tie my %valid, 'DB_File', fd_path($tmpfile), undef, oct(600), $DB_File::DB_BTREE # Does a separate open
+  or die "DB_File tie failed: $!";
 close($tmpfile); # So we can close this
 
 ### Preparation of the package lists ########################################
@@ -629,39 +676,54 @@ if($> == 0 && !$cfg->{user} && !$force) {
     die "Running $0 as root\nand no effective user has been specified. Aborting.\nPlease set the effective user in $configfile or use --force to ignore\n";
 }
 
-# Try to ensure corresponding Packages/Sources is present for each diff_Index
-# and Release for each Packages/Sources
-{
-    my %missing;
-  CHECKFILE:
-    foreach (glob('*diff_Index *{Packages,Sources}{,.gz,.bz2}')) {
-	my $file = $_;
-	if (s/\.diff_Index$/{,.gz,.bz2}/) {
-	    printmsg "Checking for $_ for $file\n";
-	}
-	elsif (s/(?:dists_[^_]+_(?:updates_)?\K(?:[^_]+_){2})?(?:Packages|Sources)(?:\.(?:bz2|gz))?$/{In,}Release/) {
-	    printmsg "checking for $_ for $file\n";
-	}
+if (!$offline) {
+    # Ensure corresponding Packages/Sources is present for each
+    # diff_Index/by-hash and Release for each Packages/Sources.
+    # Do separate passes as diff_Index or by-hash may cause download of new
+    # Packages or Sources.
+    foreach ('*diff_{Index,by-hash_*}', '*{Packages,Sources}{,.gz,.xz,.bz2}') {
+	my %missing;
+      CHECKFILE:
 	foreach (glob) {
-	    if ( -f $_ ) {
-		printmsg "Found $_\n";
-		next CHECKFILE;
+	    my $file = $_;
+	    s/\.diff_(?:Index$|(?:by-hash_(?i:MD5SUM_[0-9a-f]{32}|SHA1_[0-9a-f]{40}|SHA256_[0-9a-f]{64})))/{.xz,.gz,.bz2,}/
+	      or
+	      s/(?:dists_[^_]+_(?:updates_)?\K(?:[^_]+_){2})?(?:Packages|Sources)(?:\.(?:bz2|(?:x|g)z))?$/{In,}Release/;
+	    printmsg "checking for $_ as index for $file\n";
+	    foreach (glob) {
+		if ( -f $_ ) {
+		    printmsg "Found $_\n";
+		    next CHECKFILE;
+		}
 	    }
+	    push @{$missing{$_}}, $file;
+	}
+
+      MISSING:
+	foreach (keys %missing) {
+	    foreach (glob) {
+		my $resp;
+		get($_);
+		$resp = read_header("../headers/$_") if  -f "../headers/$_" and -f $_;
+		next MISSING if $resp && $resp->is_success;
+		printmsg "$_ not available upstream\n";
+	    }
+	    warn "$_ not available upstream. \nRemove the following files manually if they are no longer significant and you want to clean cached packages they reference:-\n";
+	    warn "  $_\n" foreach @{$missing{$_}};
+	    die "Exiting to prevent deletion of cache contents.\n";
 	}
-	$missing{(glob)[-1]} = 1;
     }
-    get($_) foreach keys %missing;
 }
 
 
 # Initially preserve the index files
-%valid = map {$_ => 1} glob('*{Release,diff_Index} *{Packages,Sources}{,.gz,.bz2}');
+%valid = map {$_ => 1} glob('*{Release,diff_Index} *{Packages,Sources}{,.gz,.xz,.bz2}');
 
 foreach my $file (keys %valid) {
 
     # Try to patch
     my $patched;
-    if($pdiff_mode && $file =~ /(?:Packages|Sources)(?:\.(?:bz2|gz))?$/) {
+    if($pdiff_mode && $file =~ /(?:Packages|Sources)(?:\.(?:bz2|(?:x|g)z))?$/) {
 	printmsg "Attempting to update $file by patching\n";
 	($patched = pdiff($file)) || printmsg "Patching failed or not possible\n";
     }
@@ -675,6 +737,16 @@ foreach my $file (keys %valid) {
 	}
     }
 
+    # Handle by-hash index files
+    if ($file =~ /InRelease$/) {
+	foreach my $candidate (@{by_hash($file)}) {
+	    if (-f $candidate) {
+		printmsg "Cached file $candidate is current by-hash file referenced in $file. Adding to list for parsing\n";
+		$valid{$candidate} = 1;
+	    }
+	}
+    }
+
     # Remove obsolete Release and Release.gpg
     if ((my $obsolete = $file) =~ s/InRelease/Release/) {
 	next unless exists $valid{$obsolete};
@@ -685,10 +757,22 @@ foreach my $file (keys %valid) {
 
 }
 
+my %parsed;
 foreach my $file (keys %valid) {
 
     printmsg "Reading: $file\n";
 
+    # Assume different compressed formats of the same file are identical
+    if ($file=~ /(^.+(?:Packages|Sources))(?:\.(?:bz2|(?:x|g)z))?$/) {
+	if ($parsed{$1}) {
+	    printmsg "Already read $parsed{$1}, so skipping $file\n";
+	    next;
+	}
+	else {
+	    $parsed{$1} = $file;
+	}
+    }
+
     if ((my ($cfh, undef) = open_lock($file)) == 2) {
 	extract_sums($file, $cfh, \%valid) || die("Error processing $file in $cfg->{cache_dir}/packages, cleanup stopped.\nRemove the file if the repository is no longer interesting and the packages pulled from it are to be removed.\n");
     }
@@ -696,6 +780,7 @@ foreach my $file (keys %valid) {
 	die "Failed to open filehandles for $file. Resolve this manually. \nExiting to prevent deletion of cache contents.\n";
     }
 }
+undef %parsed;
 
 printmsg "Found ".scalar (keys %valid)." valid file entries\n";
 #print join("\n",keys %valid);
@@ -703,23 +788,21 @@ printmsg "Found ".scalar (keys %valid)." valid file entries\n";
 # Remove old checksum data
 if ($cfg->{checksum}) {
 
-    my $dbh = db();
-
-    my $do_compact;
-    $dbh && do {
 	printmsg "Removing expired entries from checksum database\n";
 
-	my $cursor = get_cursor($dbh,1);
-	my ($filename,$data)=('','');
-	while (cursor_next($cursor, \$filename, \$data) == 0)
-	  {
-	      next if defined $valid{$filename};
-	      printmsg "Deleting checksum data for $filename\n";
-	      $cursor->c_del == 0 || warn "c_del failed: $BerkeleyDB::Error" if !$sim_mode;
-	      $do_compact = 1;
-	  }
-	db_compact($dbh) if $do_compact || $pdiff_mode;
-    };
+	my $do_compact;
+	{ # scope $cursor
+	    my $cursor = db_cursor(1);
+	    my ($filename,$data)=('','');
+	    while (cursor_next($cursor, \$filename, \$data) == 0)
+	      {
+		  next if defined $valid{$filename};
+		  printmsg "Deleting checksum data for $filename\n";
+		  $cursor->c_del == 0 || warn "c_del failed: $BerkeleyDB::Error" if !$sim_mode;
+		  $do_compact = 1;
+	      }
+	}
+	db_compact() if $do_compact || $pdiff_mode;
 }
 
 # Clean package directory
@@ -728,43 +811,48 @@ if ($cfg->{checksum}) {
     printmsg "Building source package file/version table\n";
 
     open(my $tmpfile, "+>", undef) or die $!;
-    tie my %svrhash, 'GDBM_File', fd_path($tmpfile), &GDBM_NEWDB|&GDBM_FAST, oct(600) # Does a separate open
-      or die "GDBM_File tie failed: $!";
+    chmod oct(600), $tmpfile unless (stat $tmpfile)[2] & oct(7777) ^ oct(600); # workaround PerlIO_tmpfile() incorrect mode in 5.22.1
+    tie my %svrhash, 'DB_File', fd_path($tmpfile), undef, oct(600),  $DB_File::DB_BTREE # Does a separate open
+      or die "DB_File tie failed: $!";
     close($tmpfile); # So we can close this
-    %svrhash = map {m#([-+.a-z0-9]+_(?:\d:)?[-+.~a-zA-Z0-9]+)\.dsc# && $1 => 1} keys %valid;
+    m#([-+.a-z0-9]+_(?:\d:)?[-+.~a-zA-Z0-9]+)\.dsc# and $svrhash{$1} = 1 foreach keys %valid;
 
     foreach (glob('*{,/*}')) {
 	next if -d; # Skip directories
-	if (/([-+.a-z0-9]+_(?:\d:)?[-+.~a-zA-Z0-9]+)_changelog$/ && !$svrhash{$1}) {
+	if (/([-+.a-z0-9]+_(?:\d:)?[-+.~a-zA-Z0-9]+)_changelog$/ && !$svrhash{$1}) { # Changelogs
 	    unlink $_, "../headers/$_" unless $sim_mode;
 	    printmsg "Removing expired changelog: $_ and company...\n";
-	    next;
 	}
-
-	next unless is_file_type('package', $_) || is_file_type('pdiff', get_original_url($_)); # Package and pdiff files only
-
-	if(! defined($valid{$_})) {
-	    unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
-	    printmsg "Removing file: $_ and company...\n";
+	elsif (is_file_type('installer', get_original_url($_))) { # APT/installer/by-hash files
+	    # Verify still present upstream
+	    printmsg "Validating $_\n";
+	    get($_) unless $sim_mode;
 	}
-	else {
-	    # Verify SHA1 checksum
-	    my $target_sum = hashify(\$valid{$_})->{sha1};
-	    next unless $target_sum;
-	    # print "Validating SHA1 $target_sum for $_\n";
-	    open(my $fh, '<', $_) || die "Unable to open file $_ to verify checksum: $!";
-	    flock($fh, LOCK_EX);
-	    if (is_file_type('pdiff', get_original_url($_))) { # pdiffs need decompressing
-		$fh = IO::Uncompress::AnyUncompress->new($fh)
-		  or die "Decompression failed: $AnyUncompressError\n";
-	    }
-	    if ((my $sha1 = Digest::SHA->new(1)->addfile($fh)->hexdigest) ne $target_sum) {
+	elsif (is_file_type('package', $_) or is_file_type('pdiff', get_original_url($_))) { # Package and pdiff files
+
+	    if (! defined($valid{$_})) {
 		unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
-		printmsg "Checksum mismatch ($target_sum <=> $sha1): $_, removing\n";
+		printmsg "Removing file: $_ and company...\n";
+	    } else {
+		# Verify SHA1 checksum
+		my $target_sum = hashify(\$valid{$_})->{sha1};
+		next unless $target_sum;
+		# print "Validating SHA1 $target_sum for $_\n";
+		open(my $fh, '<', $_) || die "Unable to open file $_ to verify checksum: $!";
+		flock($fh, LOCK_EX);
+		if (is_file_type('pdiff', get_original_url($_))) { # pdiffs need decompressing
+		    $fh = IO::Uncompress::AnyUncompress->new($fh)
+		      or die "Decompression failed: $AnyUncompressError\n";
+		}
+		if ((my $sha1 = Digest::SHA->new(1)->addfile($fh)->hexdigest) ne $target_sum) {
+		    unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
+		    printmsg "Checksum mismatch ($target_sum <=> $sha1): $_, removing\n";
+		}
+		# No explicit LOCK_UN: it fails with IO::Uncompress::AnyUncompress, just rely on close
+		close $fh;
 	    }
-	    # No explicit LOCK_UN: it fails with IO::Uncompress::AnyUncompress, just rely on close
-	    close $fh;
 	}
+        next;
     }
 }
 
