@@ -114,6 +114,8 @@ sub read_config {
 							   xenial
 							   yakkety
 							   zesty
+							   artful
+							   bionic
 						      )),
 		  user => $>,
 
@@ -674,8 +676,9 @@ sub get_upstream_url {
 sub get_original_url {
     my ($filename) = @_;
 
-    # Infer from filename, assume HTTP
-    return URI->new('http://' . join('/', split(/_/, $filename)));
+    # Infer from filename, beware Translation-??_??, assume HTTP
+    $filename =~ s/(?<!Translation-[a-z]{2})_/\//g;
+    return URI->new('http://' . $filename);
 }
 
 # Stores data flattened for use in tied hashes
